package ru.dbotthepony.mc.otm.network

import net.minecraft.network.FriendlyByteBuf
import net.minecraft.resources.ResourceLocation
import net.minecraft.server.level.ServerPlayer
import net.minecraft.world.entity.Entity
import net.minecraft.world.entity.player.Player
import net.minecraftforge.event.network.CustomPayloadEvent
import net.minecraftforge.network.Channel
import net.minecraftforge.network.ChannelBuilder
import net.minecraftforge.network.NetworkDirection
import net.minecraftforge.network.PacketDistributor
import net.minecraftforge.network.SimpleChannel
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kommons.util.Delegate
import ru.dbotthepony.kommons.util.getValue
import ru.dbotthepony.kommons.util.setValue
import ru.dbotthepony.mc.otm.NULLABLE_MINECRAFT_SERVER
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import java.util.concurrent.CompletableFuture
import java.util.concurrent.ConcurrentLinkedQueue
import java.util.concurrent.locks.LockSupport
import java.util.function.BiConsumer
import java.util.function.Function
import kotlin.reflect.KClass

class MNetworkContext(val sender: ServerPlayer?, packetHandled: Delegate<Boolean>, private val enqueuer: (Runnable) -> CompletableFuture<*>) {
	var packetHandled by packetHandled

	fun enqueueWork(callback: Runnable): CompletableFuture<*> {
		return enqueuer(callback)
	}
}

interface MatteryPacket {
	fun write(buff: FriendlyByteBuf)
	fun play(context: MNetworkContext)
}

abstract class MatteryNetworkChannel(val version: Int, val name: String) {
	val channel: SimpleChannel = ChannelBuilder
		.named(ResourceLocation(OverdriveThatMatters.MOD_ID, name))
		.clientAcceptedVersions { status, version -> (status == Channel.VersionTest.Status.MISSING || status == Channel.VersionTest.Status.VANILLA) || version == this.version }
		.serverAcceptedVersions { status, version -> status == Channel.VersionTest.Status.PRESENT || version == this.version }
		.networkProtocolVersion(version)
		.simpleChannel()

	fun sendToServer(packet: Any) = channel.send(packet, PacketDistributor.SERVER.noArg())

	fun send(ply: Player, packet: Any) {
		if (ply is ServerPlayer) {
			channel.send(packet, PacketDistributor.PLAYER.with(ply))
		}
	}

	fun sendTracking(entity: Entity, packet: Any) {
		if ((NULLABLE_MINECRAFT_SERVER?.playerCount ?: 0) <= 0) {
			return
		}

		channel.send(packet, PacketDistributor.TRACKING_ENTITY.with(entity))
	}

	fun sendTrackingAndSelf(entity: Entity, packet: Any) {
		if ((NULLABLE_MINECRAFT_SERVER?.playerCount ?: 0) <= 0) {
			return
		}

		channel.send(packet, PacketDistributor.TRACKING_ENTITY_AND_SELF.with(entity))
	}

	fun send(distributor: PacketDistributor.PacketTarget, packet: Any) {
		channel.send(packet, distributor)
	}

	private var nextNetworkPacketID = 0

	fun <T> add(
		packetClass: Class<T>,
		writer: BiConsumer<T, FriendlyByteBuf>,
		reader: Function<FriendlyByteBuf, T>,
		handler: (T, MNetworkContext) -> Unit,
		direction: NetworkDirection? = null,
		handleOnMainThread: Boolean = true,
	) {
		if (nextNetworkPacketID >= 256) {
			throw IndexOutOfBoundsException("Network message ID overflow!")
		}

		val builder = channel.messageBuilder(packetClass, direction)
		val bridgeHandler = BiConsumer<T, CustomPayloadEvent.Context> { a, b -> handler(a, MNetworkContext(b.sender, Delegate.Of({ b.packetHandled }, { b.packetHandled = it }), b::enqueueWork)) }

		if (handleOnMainThread) {
			builder.consumerMainThread(bridgeHandler)
		} else {
			builder.consumerNetworkThread(bridgeHandler)
		}

		builder.encoder(writer)
		builder.decoder(reader)
		builder.add()
	}

	fun <T : MatteryPacket> add(
		packetClass: Class<T>,
		reader: Function<FriendlyByteBuf, T>,
		direction: NetworkDirection? = null,
		handleOnMainThread: Boolean = true,
	) {
		add(packetClass, MatteryPacket::write, reader, MatteryPacket::play, direction, handleOnMainThread)
	}

	fun <T : MatteryPacket> add(
		packetClass: KClass<T>,
		reader: Function<FriendlyByteBuf, T>,
		direction: NetworkDirection? = null,
		handleOnMainThread: Boolean = true,
	) {
		add(packetClass.java, MatteryPacket::write, reader, MatteryPacket::play, direction, handleOnMainThread)
	}
}
